from logging import config


class LoggingConfiguration:
    def __init__(self):
        pass

    @staticmethod
    def apply_config(log_level):
        validated_log_level = log_level.upper()
        if validated_log_level not in ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]:
            print(f"Invalid log level {log_level}, falling back to INFO")
            validated_log_level = "INFO"

        log_config = {
            "version": 1,
            "root": {"handlers": ["console"], "level": validated_log_level},
            "handlers": {
                "console": {"formatter": "std_out", "class": "logging.StreamHandler", "level": validated_log_level}
            },
            "formatters": {
                "std_out": {
                    "format": "%(asctime)s : %(levelname)s : %(message)s",
                    "datefmt": "%d-%m-%Y %I:%M:%S",
                }
            },
        }
        config.dictConfig(log_config)



import logging
import time
from concurrent.futures import ThreadPoolExecutor

from ai.h2o.featurestore.api.v1 import CoreService_pb2 as pb

from .commons.spark_utils import SparkUtils
from .job_info import JobInfo
from .utils import Utils


class RetrieveAsLinksCommon:
    def _get_job(self, job_id):
        return self._stub.GetJob(job_id)

    def _get_retrieve_links_output(self, job_id):
        return self._stub.GetRetrieveAsLinksJobOutput(job_id)

    def download(self, output_dir=None):
        """Downloads files to the specified directory location.

        Args:
            output_dir: (str) A directory location as string. Default is None.

        Returns:
            str: A directory path where the files are downloaded.

        Typical usage example:

            retrieve_job = client.jobs.get(<job id>)
            dir = retrieve_job.download()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/jobs_api.html#how-to-download-using-retrievejob
        """
        self._process_lazy_ingest_task()
        if not self._job_id:
            self._job_id = self._start_retrieve_links_job()

        info = JobInfo(self._stub, self._job_id)
        while not self._get_job(self._job_id).done:
            info.show_progress()
            time.sleep(2)
        info.show_progress()  # there is possibility that some progress was pushed before finishing job
        retrieve_as_links_response = self._get_retrieve_links_output(self._job_id)
        return Utils.download_files(output_dir, retrieve_as_links_response.download_links)

    def download_async(self, output_dir=None):
        """Downloads files asynchronously to the specified directory location.

        Args:
            output_dir: (str) A directory location as string. Default is None.

        Returns:
            DownloadFuture: Represents a job.

            It contains is_done and get_status methods. For example:

            future.is_done() -> check whether job has finished or not
            future.get_status() -> > return job's status

        Typical usage example:

            retrieve_job = client.jobs.get(<job id>)
            future = retrieve_job.download_async()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/jobs_api.html#how-to-download-using-retrievejob
        """
        future = self._thread_pool.submit(self.download, output_dir)
        return DownloadFuture(future)

    def _start_retrieve_links_job(self):
        request = self._create_retrieve_request()
        return self._stub.StartRetrieveAsLinksJob(request)

    def _create_retrieve_request(self, spark_session=None):
        session_id = ""
        if spark_session is not None:
            session_id = spark_session.conf.get("ai.h2o.featurestore.sessionId", "")
        request = pb.RetrieveRequest()
        request.feature_set_id = self._feature_set.id
        request.feature_set_version = self._feature_set.version
        request.session_id = session_id
        if self._start_date_time is not None:
            request.start_date_time = self._start_date_time
        if self._end_date_time is not None:
            request.end_date_time = self._end_date_time
        request.ingest_id = self._ingest_id
        return request

    def _process_lazy_ingest_task(self):
        request = pb.LazyIngestRequest(
            feature_set_id=self._feature_set.id, feature_set_version=self._feature_set.version
        )
        response = self._stub.StartLazyIngestTask(request)
        if response.job_id.job_id:
            logging.info(
                "No previous ingestion found. Lazy ingest task will start. "
                "New minor version will be created and retrieved from."
            )
            info = JobInfo(self._stub, response.job_id)
            while not self._get_job(response.job_id).done:
                info.show_progress()
                time.sleep(2)
            info.show_progress()  # there is possibility that some progress was pushed before finishing job
            feature_set_response = self._stub.GetIngestJobOutput(response.job_id).feature_set
            self._feature_set.version = feature_set_response.version


class RetrieveHolder(RetrieveAsLinksCommon):
    def __init__(self, stub, feature_set, start_date_time, end_date_time, ingest_id):
        self._stub = stub
        self._thread_pool = ThreadPoolExecutor(5)
        self._feature_set = feature_set
        self._start_date_time = start_date_time
        self._end_date_time = end_date_time
        self._retrieve_as_spark_response = None
        self._job_id = None
        self._ingest_id = ingest_id

    def download(self, output_dir=None):
        """Downloads files to the specified directory location.

        Args:
            output_dir: (str) A directory location as string. Default is None.

        Returns:
            str: A directory path where the files are downloaded.

        Typical usage example:

            ref = feature_set.retrieve()
            dir = ref.download()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#downloading-the-files-from-feature-store
        """
        return super(RetrieveHolder, self).download(output_dir)

    def download_async(self, output_dir=None):
        """Downloads files asynchronously to the specified directory location.

        Args:
            output_dir: (str) A directory location as string. Default is None.

        Returns:
            DownloadFuture: Represents a job.

            It contains is_done and get_status methods. For example:

            future.is_done() -> check whether job has finished or not
            future.get_status() -> return job's status

        Typical usage example:

            ref = feature_set.retrieve()
            future = ref.download_async()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#downloading-the-files-from-feature-store
        """
        return super(RetrieveHolder, self).download_async(output_dir)

    def as_spark_frame(self, spark_session):
        """Returns a spark data frame.

        Generates a data frame of the retrieved data using spark session.

        Args:
            spark_session: (SparkSession) Represents a spark session.

        Returns:
            DataFrame: Represents a spark data frame.

            The frame is made up of columns as features and rows as records.
            Rows contain data within retrieve scope (filtered).

        Typical usage example:

            ref = feature_set.retrieve()
            data_frame = ref.as_spark_frame(spark_session)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#obtaining-data-as-a-spark-frame
        """
        from pyspark.sql.functions import col, from_utc_timestamp, lit, to_timestamp, unix_timestamp
        from pyspark.sql.types import TimestampType

        self._process_lazy_ingest_task()
        if self._retrieve_as_spark_response is None:
            request = self._create_retrieve_request(spark_session)
            self._retrieve_as_spark_response = self._stub.RetrieveAsSpark(request)
        resp = self._retrieve_as_spark_response
        spark_session.conf.set("ai.h2o.featurestore.sessionId", resp.session_id)
        SparkUtils.configure_user_spark(spark_session)
        for k, v in resp.options.items():
            spark_session.conf.set(k, v)
        if self._retrieve_as_spark_response.delta_version == -1:
            df = spark_session.read.format("delta").load(resp.cache_path)
        else:
            df = (
                spark_session.read.format("delta")
                .option("versionAsOf", self._retrieve_as_spark_response.delta_version)
                .load(resp.cache_path)
            )
        retrieve_scope = resp.retrieve_scope
        start_timestamp = retrieve_scope.start_date_time.seconds
        end_timestamp = retrieve_scope.end_date_time.seconds
        timestamp_col = "timestamp_" + str(round(time.time() * 1000))
        if self._ingest_id:
            output_df = df.filter(col("ingest_id") == lit(self._ingest_id))
        elif self._feature_set.time_travel_column:
            output_df = (
                df.withColumn(
                    timestamp_col,
                    unix_timestamp(
                        from_utc_timestamp(
                            to_timestamp(
                                col("`" + self._feature_set.time_travel_column + "`"),
                                self._feature_set.time_travel_column_format,
                            ),
                            spark_session.conf.get("spark.sql.session.timeZone"),
                        )
                    ),
                )
                .filter(col(timestamp_col) <= end_timestamp)
                .filter(col(timestamp_col) >= start_timestamp)
            ).drop(timestamp_col)
        else:
            output_df = (
                df.filter(col("time_travel_column_auto_generated") <= end_timestamp)
                .filter(col("time_travel_column_auto_generated") >= start_timestamp)
                .withColumn(
                    "time_travel_column_auto_generated", col("time_travel_column_auto_generated").cast(TimestampType())
                )
            )
        internal_columns = ["ingest_id"] + [column for column in df.columns if column.startswith("__")]
        return output_df.drop(*internal_columns)


class DownloadFuture:
    def __init__(self, future):
        self._future = future
        self._result = None

    def is_done(self) -> bool:
        """Checks whether job has finished or not."""
        return self._future.done()

    def get_result(self):
        """Returns job's final result.

        Returns:
            Schema: A resultant schema with feature names and data types.

            For example:

            id INT, text STRING, label DOUBLE, state STRING, date TIMESTAMP

            Above result is the output of asynchronous job for schema extraction.

        Typical example:

            job = client.extract_schema_from_source_async(source)
            schema.get_result()

        Raises:
            Exception: Job has not finished yet!
        """
        if not self._result:
            if not self.is_done():
                raise Exception("Job has not finished yet!")
            self._result = self._future.result()
        return self._result



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..utils import Utils
from .feature_set import FeatureSet


class PinnedFeatureSet:
    def __init__(self, stub, pinned_feature_set):
        self._stub = stub
        self._pinned_feature_set = pinned_feature_set

    @property
    def name(self):
        return self._pinned_feature_set.feature_set_name

    @property
    def description(self):
        return self._pinned_feature_set.feature_set_description

    @property
    def updated_at(self):
        return Utils.timestamp_to_string(self._pinned_feature_set.updated_at)

    @property
    def pinned_at(self):
        return Utils.timestamp_to_string(self._pinned_feature_set.pinned_at)

    def get_feature_set(self):
        request = pb.FeatureSetRequestId(feature_set_id=self._pinned_feature_set.feature_set_id)
        response = self._stub.GetFeatureSetById(request)
        return FeatureSet(self._stub, response.feature_set)

    def __repr__(self):
        return Utils.pretty_print_proto(self._pinned_feature_set)



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..access_type import AccessType
from ..utils import Utils
from .user import User


class UserWithPermission:
    def __init__(self, user_with_permission):
        self._user_with_permission = user_with_permission

    @property
    def user(self):
        return User(self._user_with_permission.user)

    @property
    def access_type(self):
        return AccessType.from_proto_permission(self._user_with_permission.resource_permission.permission)

    @property
    def resource_type(self):
        return pb.ResourceType.Name(self._user_with_permission.resource_permission.resource_type)

    def __repr__(self):
        return Utils.pretty_print_proto(self._user_with_permission)

    def __str__(self):
        return (
            f"User\n{Utils.output_indent_spacing(str(self.user), '  ')}"
            f"Access Type   : {self.access_type}\n"
            f"Resource Type : {self.resource_type}\n"
        )



import re
from typing import Optional

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..entities.feature_set import FeatureSet
from ..feature_set_flow import FeatureSetFlow
from ..schema import Schema
from ..utils import Utils


class FeatureSets:
    def __init__(self, stub, project):
        self._project = project
        self._stub = stub

    def register(
        self,
        schema,
        feature_set_name,
        description="",
        primary_key=None,
        time_travel_column=None,
        time_travel_column_format="yyyy-MM-dd HH:mm:ss",
        secret=False,
        partition_by=None,
        time_travel_column_as_partition=False,
        flow: Optional[FeatureSetFlow] = None,
    ):
        """Create a new feature set.

        Args:
            schema: (Schema) A schema that contains feature columns and data types.
            feature_set_name: (str) A name for the feature set.
            description: (str) A description about the feature set.
            primary_key: (str | list[str]) A key / keys for a feature column name.
            time_travel_column: (str) A feature column in a schema.
            time_travel_column_format: (str) Format for time travel column.
            secret: (bool) Object controls visibility of the feature set.
            partition_by: (list[str]) Object represents a list of String.
            time_travel_column_as_partition: (bool) Feature Store uses time travel column for data partitioning.
            flow: (FeatureSetFlow) Feature store uses flow to control where data lives (offline, online or both)

        Returns:
            FeatureSet: A new feature set with specified attributes.

        Typical example:
            project.feature_sets.register(schema, "feature_set_name", description="", primary_key=None,
              time_travel_column=None, time_travel_column_format="yyyy-MM-dd HH:mm:ss", secret=False,
              partition_by=None, time_travel_column_as_partition=False)

        Raises:
            ValueError: Parameter schema should be of type Schema.

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#registering-a-feature-set
        """
        if not isinstance(schema, Schema):
            raise ValueError("Parameter `schema` should be of type `featurestore.core.schema.Schema`")
        request = pb.RegisterFeatureSetRequest()
        request.schema.extend(schema._to_proto_schema())
        request.project.CopyFrom(self._project)
        if schema.derivation is not None:
            schema.derivation.transformation._initialize(self._stub)
            request.derived_from.CopyFrom(schema.derivation._to_proto())
        if primary_key is not None:
            if isinstance(primary_key, str):
                request.primary_key.append(primary_key)
            else:
                request.primary_key.extend(primary_key)
        if time_travel_column is not None:
            request.time_travel_column = time_travel_column
        request.secret = secret
        request.description = description
        if partition_by is not None:
            request.partition_by.extend(partition_by)
        request.time_travel_column_as_partition = time_travel_column_as_partition
        request.time_travel_column_format = time_travel_column_format
        request.feature_set_name = feature_set_name
        if flow is not None:
            request.flow = flow.name
        response = self._stub.RegisterFeatureSet(request)
        self._reload_project()
        return FeatureSet(self._stub, response.feature_set)

    def get(self, feature_set_name, version=None):
        """Obtain an existing feature set.

        Args:
            feature_set_name: (str) A feature set name.
            version: (str) A specific version of feature set with format as "major.minor".

        Returns:
            FeatureSet: An existing feature set.

        Typical example:
            fs = project.feature_sets.get("feature_set_name", version=None)

        Raises:
            Exception: Version parameter must be in a format "major.minor".

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#obtaining-a-feature-set
        """
        request = pb.GetFeatureSetRequest()
        request.project_id = self._project.id
        request.feature_set_name = feature_set_name
        if version is not None:
            if not re.search(r"^\d+\.\d+$", str(version)):
                raise Exception('Version parameter must be in a format "major.minor".')
            request.version = str(version)
        response = self._stub.GetFeatureSet(request)
        return FeatureSet(self._stub, response.feature_set)

    def list(self):
        """Return a generator which obtains the feature sets lazily.

        Returns:
            Iterable[FeatureSet]: A generator iterator object consists of feature sets.

        Typical example:
            project.feature_sets.list()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_api.html#listing-feature-sets-within-a-project
        """
        request = pb.ListFeatureSetsPageRequest()
        request.project_names.extend([self._project.name])
        while request:
            response = self._stub.ListFeatureSetsPage(request)
            if response.next_page_token:
                request.page_token = response.next_page_token
            else:
                request = None
            for feature_set in response.feature_sets:
                yield FeatureSet(self._stub, feature_set)

    def __repr__(self):
        return Utils.pretty_print_proto(self._project)

    def _reload_project(self):
        request = pb.GetProjectRequest()
        request.project_name = self._project.name
        response = self._stub.GetProject(request)
        self._project = response.project
